#lang plai


#|Definición de las expresiones|#
(define-type MinTyRacket
  [NULL (tipo tipo_elem?)]
  [FST(ls MinTyRacket?)]
  [ISEMPTY? (ls MinTyRacket?)]
  [CONST(elem MinTyRacket?) (ls MinTyRacket?)]
  [RST (ls MinTyRacket?)]
  [ERROR]
  [OP (op symbol?) (lhs MinTyRacket?) (rhs MinTyRacket?)]
  [VAR (nombre symbol?)]
  [LAMBDA (argus (listof symbol?)) (types (listof symbol?))  (body MinTyRacket?)]
  [IF (cond MinTyRacket?) (true MinTyRacket?) (else MinTyRacket?)]
  [APP (f MinTyRacket?) (args (listof MinTyRacket?))]
  [FIX (varlig symbol?) (nombre MinTyRacket?) (cuerpolig MinTyRacket?)]
  [true_]
  [false_]
  [LIST (ls  list?)]
  [bool (n  boolean?)]
  [num (n  integer?)]
  [flot (n  flonum?)]
  [doub (n  real?)])


#|Definición de la operación aritmética -|#
(define (resta c d)  
  ((getSupertipo c d) (- (typeV-n c) (typeV-n d))))


#|Definición de la operación aritmética división|#
(define (division m n)   
  (cond 
    ([equal? (typeV-n n) 0] (ERROR)) 
    (else (cond
             ([integer? (/ (typeV-n m) (typeV-n n))]    
             (integerV (/ (typeV-n m) (typeV-n n))))
             ([real? (/ (typeV-n m) (typeV-n n))] 
             (doubleV (/ (typeV-n m) (typeV-n n))))
             ))))
           

#|Definición de la operación booleana menor que|#
(define (menorque a b)  (booleanV (< (typeV-n a) (typeV-n b))))


#|Definición de la operación booleana menor que|#
(define (mayorque a b)  (booleanV (> (typeV-n a) (typeV-n b))))


#|Definición de la operación aritmética multiplicación|#
(define (multiplicacion j k)   ((getSupertipo j k) (* (typeV-n j) (typeV-n k))))   


#|Definición de la operación aritmética suma|#
(define (suma a b)  ((getSupertipo a b) (+ (typeV-n a) (typeV-n b))))


#|Definición de la operación auxiliar de igualdad|#
(define (igualdad a b)  
  (booleanV  (equal? (getComponente a) (getComponente b) ) )
 )

#|Definición de la operación que obtiene los componentes de los valores|#
(define (getComponente expr)
     (cond 
       ((MinTyRacket? expr)   (MintyRacket-n expr) )
       ((MinTyRacket_Valores? expr)       (typeV-n expr) )
       (else (append "La operación no puede realizarse con tales argumentos"))
       ))



#|Definición del conjunto de operaciones binarias|#
(define opbinarias (list 
                    (list '+ suma)
                    (list '< menorque)
                    (list '> mayorque)
                    (list '== igualdad)
                    (list '- resta)
                    (list '* multiplicacion)
                    (list '/ division)))


#|Definición de los valores|#
(define-type MinTyRacket_Valores
  [integerV (n integer?)]
  [floatV (n flonum?)]
  [doubleV (n real?)]
  [closureV (params (listof symbol?))
            (body MinTyRacket?)
            (env Env?)]
  [booleanV (n boolean?)] 
  [listV  (ls list?)]
  [consV (cs cons?)])


#|Procedimiento que comprueba cuando se trata de un valor|#
(define (isInMinTyRacket_Valores? lo)
  (and (box? lo)
       (MinTyRacket_Valores? (unbox lo))))


#|Definición del ambiente de valores|#
(define-type Env 
  [mtEnv]
  [anEnv (name symbol?) (value MinTyRacket_Valores?) (env Env?)]
  [aRecSub (name symbol?)
           (value isInMinTyRacket_Valores?)
           (env Env?)])


#|Definición de los tipos elementales|#
(define-type tipo_elem
  [int]
  [float]
  [double]
  [booln]
  [Min]
  [Max]
  )


#|Definición de los tipos|#
(define-type tipo
  [lista (elementos tipo_elem?)]
  [flecha (antec tipo_elem?) (consec tipo_elem?)])


#|Definición de valores simples|#
(define (typeV-n value)
  (type-case MinTyRacket_Valores value
    [listV (n) n]
    [integerV (n) n]
    [floatV (n) n]
    [doubleV (n) n]
    [booleanV (n) n]
    [consV (n) n]
    {else (append "El tipo no ha sido definido")} ))
  

(define (MintyRacket-n exp)
  (type-case MinTyRacket exp
    [num (n) n]
    [flot (n) n]
    [doub (n) n]
    [bool (n) n]
    {else (append "La expresión no permite obtener sus componentes")} ))


#|Definición de tablas hash con los valores asociados|#
(define htV (make-hash))
  (hash-set! htV integerV integerV)
  (hash-set! htV floatV floatV)
  (hash-set! htV doubleV doubleV)
  (hash-set! htV booleanV booleanV)
  (hash-set! htV integerV-n integerV-n)
  (hash-set! htV floatV-n floatV-n)
  (hash-set! htV doubleV-n doubleV-n)
  (hash-set! htV booleanV-n booleanV-n)
  (hash-set! htV float floatV)
  (hash-set! htV double doubleV)
  (hash-set! htV booln booleanV)
  (hash-set! htV int integerV)
  (hash-set! htV list listV)
  (hash-set! htV #t true)
  (hash-set! htV #f false)


(define (exprV expresionV)
  (cond 
    [(integerV? expresionV)(hash-ref htV integerV)]
    [(floatV? expresionV)(hash-ref htV floatV)]
    [(doubleV? expresionV)(hash-ref htV doubleV)]
    [(booleanV? expresionV)(hash-ref htV booleanV)]))

(define (exp-V expresionV)
  (cond 
    [(integerV? expresionV)(hash-ref htV integerV-n)]
    [(floatV? expresionV)(hash-ref htV floatV-n)]
    [(doubleV? expresionV)(hash-ref htV doubleV-n)]
    [(booleanV? expresionV)(hash-ref htV booleanV-n)]
    ))

(define htVJ (make-hash))
(hash-set! htVJ integerV 0)
(hash-set! htVJ floatV 1)
(hash-set! htVJ doubleV 2)
(hash-set! htVJ booleanV 3)


#|Definición de operaciones que generan la jerarquía de valores|#
(define (jerarquiaValores a b)
  (if (< (hash-ref htVJ a) (hash-ref htVJ b) 
      ) 
      (hash-ref htV b) (hash-ref htV a)))


(define (gettypeV value)
  (type-case MinTyRacket_Valores value
    [integerV (n) (hash-ref htV integerV)]
    [floatV (n) (hash-ref htV floatV)]
    [doubleV (n) (hash-ref htV doubleV)]
    {else (append "El tipo no ha sido definido")} ))


(define (getSupertipo a b)
  (jerarquiaValores (gettypeV a)(gettypeV b)))


#|Definición de asocia un valor aritmético a una expresión de MinTyRacket|#
(define (asociarValorMTR_Tipo value)
 (type-case tipo_elem value
    [int () (hash-ref htV int)]
    [float () (hash-ref htV float)]
    [double () (hash-ref htV double)]
    {else (append "El tipo no ha sido definido")} ))


#|Definición de asocia un valor a una expresión de MinTyRacket|#
(define (getValorMTR evalResult tipoProg)
  (cond 
    [(ERROR? evalResult) evalResult]
    [(or (float? tipoProg) (double? tipoProg))
     (integerV? evalResult)
     ((hash-ref htV (asociarValorMTR_Tipo tipoProg))
      (+(typeV-n evalResult) 0.0))]
    [(floatV? evalResult)
     ((hash-ref htV (asociarValorMTR_Tipo tipoProg)) evalResult)]
    [(doubleV? evalResult)
     ((hash-ref htV (asociarValorMTR_Tipo tipoProg)) evalResult)]
    [(booleanV? evalResult)
     ((hash-ref htV booleanV)evalResult)]
    [(boolean? evalResult)
     ((hash-ref htV booleanV)(hash-ref htV evalResult))]
    [(list? evalResult)
     (listV (construirList getValorMTR evalResult (getTipoProgList tipoProg)))]
    [(list? tipoProg)(getValorMTR evalResult (second tipoProg))]
    [else evalResult]
    ))


(define (construirList fn ls getTipoProgListResul)
  (if (null? ls)
      '()
      (cons (fn (car ls) getTipoProgListResul) 
            (construirList fn (cdr ls) getTipoProgListResul))))


#|Definición de asocia un valor listV a una expresión de MinTyRacket|#
(define (getTipoProgList tipoProg)
  (type-case tipo tipoProg
    [lista (n) 
       (cond
         [(integer? n) (int)]
         [(float? n) (float)]
         [(double? n) (double)]
         [else "no existe asociación al tipo indicado por el programador"]
         )    
     ]
    {else (append "El tipo no ha sido definido")} ))

