#lang plai




(require "DefinicionesMTR.rkt")




#|Regresa la lista de identificadores de acotamiento|#
(define Arg
  (lambda (ls) 
    (if (null? ls) '()
(cons (car ls) '() ))))


#|Regresa la lista de valores de ligado|#
(define val
  (lambda (ls) 
    (if (null? ls) '()
        (cons (third ls) '()  ))))


#|Regresa la lista de tipos|#
(define types
  (lambda (ls) 
    (if (null? ls) '()
    (cons (second  ls) '() ))))


#|Obtiene la representación abstracta de cada elemento de una lista|#    
(define identidadAbstracta
  (lambda (lista)   
    (if (= (length lista) 0)
        null
        ( append  (identidadAbstracta (remove (last lista) lista) )    (list (parse(last lista))))))
  )



#|Función que traduce expresiones escritas en sintáxis concreta a sintáxis abstracta|#
(define (parse sexp)
  (cond
    
    ((boolean? sexp) (bool sexp))
    ((integer? sexp) (num sexp))
    ((flonum? sexp) (flot sexp))
    ((real? sexp) (double sexp))
    ((equal?  sexp 'true) (parse true))
    ((equal? sexp 'false) (parse false))
    ((symbol? sexp) (VAR sexp))  
    ((equal? (first sexp) 'null)  (NULL (parseTiposAuxiliar (first(second sexp))) ))
    ((and (not(equal? sexp 'cons)) (and (symbol? sexp)  
                                    (and (not(equal? sexp 'false_))))) (VAR sexp))
    ((equal? (first sexp) 'fst) (FST  (parse (second sexp))))
    ((equal? (first sexp) 'list)  (LIST (identidadAbstracta (rest sexp))))
    ((equal? (first sexp) 'rst) (RST  (parse (second sexp))))
    ((equal? (first sexp) 'isempty?) (ISEMPTY?  (parse (second sexp))))
    ((equal? (first sexp) 'cons) (CONST (parse (second sexp)) (parse (third sexp)) ))
    ((equal?(first sexp) 'error ) (ERROR))
    ((equal? (first sexp) 'let)   
                         (APP (LAMBDA (Arg (second sexp))  (types (second sexp)) (parse (third sexp)))
                              (map parse(rest (append '('a) (val (second sexp)))))))  
    ((equal? (first sexp) 'if) 
                         (IF (parse (second sexp)) (parse (third sexp)) (parse (fourth sexp))))
    ((equal? (first sexp) 'lambda)       
                         (LAMBDA (Arg (second sexp) )  (types (second sexp)) (parse (third sexp))))
    ((equal? (first sexp) '+) (OP '+ (parse (cadr sexp)) (parse (caddr sexp))))
    ((equal? (first sexp) '==) (OP '== (parse (cadr sexp)) (parse (caddr sexp))))
    ((equal? (first sexp) '<) (OP '< (parse (cadr sexp)) (parse (caddr sexp))))
    ((equal? (first sexp) '>) (OP '> (parse (cadr sexp)) (parse (caddr sexp))))
    ((equal? (first sexp) 'fix) (FIX (first (second sexp))  
                                  (parse (second (second sexp)))
                                  (parse (third sexp)))) 
    ((equal? (first sexp)'/) (OP '/ (parse (cadr sexp)) (parse (caddr sexp))))
    ((equal? (first sexp) '-) (OP '- (parse (cadr sexp)) (parse (caddr sexp))))
    ((equal? (first sexp)'*) (OP '* (parse (cadr sexp)) (parse (caddr sexp))))
    (else  (APP (parse (first sexp)) (map parse (list (second sexp)))))))
        

          
#|Función que traduce tipos escritos en sintaxis concreta a sintaxis abstracta|# 
(define (parseTipos sexp)
  (cond
    ((boolean? sexp) (bool sexp))
    ((integer? sexp) (num sexp))
    ((flonum? sexp) (flot sexp))
    ((real? sexp) (double sexp))
    ((equal?  sexp 'true) (parseTiposAuxiliar 'true_))
    ((equal? sexp 'false) (parseTiposAuxiliar 'false_))
    ((equal? (first sexp) 'null)  (lista (parseTiposAuxiliar(second sexp))))
    ((and (not(equal? sexp 'cons_)) (and (symbol? sexp)  (and (not(equal? sexp 'false_)) 
                                                              ))) (VAR sexp))
    ((equal? (first sexp) 'fst) (parseTiposAuxiliar(third sexp)))
    ((equal? (first sexp) 'list)  (LIST (identidadAbstracta (rest sexp))))
    ((equal? (first sexp) 'rst) (parseTiposAuxiliar (third sexp)))
    ((equal? (first sexp) 'isempty?) (parseTiposAuxiliar (third sexp)))
    ((equal? (first sexp) 'cons) (parseTiposAuxiliar (fourth sexp)) )
    ((equal? (first sexp) 'let)   
     (parseTiposAuxiliar (fourth sexp))) 
    ((equal? (first sexp) 'if) (parseTiposAuxiliar(fifth sexp)))
    ((equal? (first sexp) 'lambda)     
     (LAMBDA (Arg (second sexp) )  (types (second sexp)) (parse (third sexp))))
    ((equal? (first sexp) '+) (parseTiposAuxiliar (fourth sexp)))
    [(equal? (first sexp) 'fix) (cons  
                                 (parseTiposAuxiliar (third (second sexp)))
                                 (list (parseTiposAuxiliar (fourth sexp))))]     
    ((equal? (first sexp)'/)(parseTiposAuxiliar (fourth sexp)))
    ((equal? (first sexp) '-) (parseTiposAuxiliar (fourth sexp)))
    ((equal? (first sexp)'*) (parseTiposAuxiliar (fourth sexp)))
    (else (parseTiposAuxiliar(third sexp)))))
     


#|Función que ayuda a transformar los tipos obtenidos en sintaxis concreta a sintaxis abstracta|#
(define (parseTiposAuxiliar n)
  (cond
    [ (equal? n 'boolean) (booln)]
    [ (and (list? n)  (equal? (first n) 'boolean))  (booln)]
    [ (equal? n 'float) (float)]
    [ (equal? n 'integer) (int)]
    [ (equal? n 'double) (double)]
    [ (equal? n 'Min) (Min)] 
    [ (equal? n 'Max) (Max)]
    [ (equal? n 'true) (booln)]
    [ (equal? n 'false) (booln)]
    [ (equal? n 'null) 'list_]
    [ (and (> (length n) 1) (and (list? n)  (equal? (second n) '->)))  
                                 (flecha (parseTiposAuxiliar (first n))
                                         (parseTiposAuxiliar (third n)))]
    [ (and (list? n)  (equal? (first n) 'List))  (lista (parseTiposAuxiliar (second n)))]
    [ (equal? (first n) 'lambda) 'fun]
    [ (and (list? n)  (member (first n) '(+ - * /))) (parseTiposAuxiliar (fourth n))]
    [ (and (> (length n) 1)  (nand (equal? (first n) 'float) (equal? (second n) '->)))  (float)]
    [ (and (list? n)  (equal? (first n) 'integer))  (int)]
    [ (and (list? n)  (equal? (first n) 'float))  (float)]
    [ (and (list? n)  (equal? (first n) 'double))  (double)]
    [ (and (list? n)  (equal? (first n) 'Max))  (Max)]
    [ (and (list? n)  (equal? (first n) 'Min))  (Min)]
    [ (and (list? n)  (equal? (first n) 'boolean))  (booln)]
    [else (error "Tipo no definido en MinTyRacket")] ))
     

#|Función que abstrae una expresion escrita en sintaxis concreta a sintaxis abstracta|#
(define (abs expr)
  (parse expr))
          