#lang plai





(require "DefinicionesMTR.rkt")
(require "ParserMTR.rkt")
(require "EvaluadorMTR.rkt")





#|Procedimiento que extiende el ambiente para el verificador de tipos|#
(define (extender_env params argu_list foo_env arg_env)
  (cond
    ((and (empty? params)
          (empty? argu_list))
     foo_env)
    ((or (and (empty? params)
              (not (empty? argu_list)))
         (and (not (empty? params))
              (empty? argu_list)))
     (error "lista de parametros y lista de argumentos no son de la misma longitud"))
    (else
     (extender_env (cdr params)
                   (cdr argu_list)
                   (anEnvTipos (car params)
                               (parseTiposAuxiliar (car argu_list))
                               foo_env)
                   arg_env))))




#|Procedimiento que implementa el algoritmo de tipificación|#
(define (verificador expr ds) 
  
  (type-case MinTyRacket expr
    [ERROR () (Min)]
    [num (n) (int)]
    [flot (n) (float)]
    [doub (n) (double)]
    [bool (n) (booln)]
    [LIST (ls) 
           (cond
             [((listof num?) ls) (lista (int))]
             [((listof flot?) ls) (lista (float))]
             [((listof doub?) ls) (lista (double))]
             [((listof bool?) ls) (lista (booln))]
             )]
    [ISEMPTY? (ls) 
              (let*[
                    (tipoT1 (verificador  ls  ds)) 
                    ]
                (type-case tipo tipoT1
                  [lista (tipoLista) (booln)]
                  [else "Tipo lista esperado"]))]
    [CONST(elem ls)
          (let*[
                (tipoT1 (verificador  elem  ds))
                (tipoT2 (verificador  ls  ds))
                ]
            (
             if (equal? (lista-elementos tipoT2)  tipoT1)
                tipoT2 
                (error "La lista recibida como parámetro no corresponde a las reglas de tipificación")
                ))]
    [FST(ls)
        (let*[
              (tipoT1 (verificador  ls  ds))
              ]
          (lista-elementos tipoT1)    
          )]
    [RST (ls)
         (let*[
               (tipoT1 (verificador  ls  ds))
               ]
           (lista (lista-elementos tipoT1) )
           )]
    [true_ () (booln)]
    [NULL (tipo) (lista tipo)]
    [false_ () (booln)]
    [OP (op IZQ DER)
           (cond
             [(equal? '== op) (booln)]
             [(equal? '> op) (booln)]
             [ 
              (let*
                  ([tipo_OP_IZQ (verificador IZQ ds)]
                   [tipo_OP_DER (verificador DER ds)])
                (cond  
                  [(equal? (ERROR) tipo_OP_DER) (Min)]
                  [(equal? (booln) tipo_OP_IZQ) 
                   (error "Error de tipos, se esperaba un número, se ha dado una expresión de tipo boolean")]
                  [(equal? (booln) tipo_OP_DER)
                   (error "Error de tipos, se esperaba un número, se ha dado una expresión de tipo boolean")] 
                  [else(JOIN tipo_OP_IZQ tipo_OP_DER)])
                )])]
    [VAR (v) (lookupTipo v ds)]
    [IF (cond true else)
         (if (booln? (verificador cond ds))
             (let* 
                 ([tipo_rama_true (verificador true ds)]
                  [tipo_rama_else (verificador else ds)]
                  )
               (JOIN tipo_rama_true tipo_rama_else))
             (error "La condicion no es de tipo booln:  " (eval cond ds)))]
    [LAMBDA (paramlist types  body)
              (let* 
                  ([env_extendido (extender_env paramlist types ds ds)]
                     [tipo_body (verificador body env_extendido)]) 
                  (flecha (parseTiposAuxiliar (car types)) tipo_body))]
    [FIX (varlig named cuerpolig)
      (type-case MinTyRacket named 
        [LAMBDA (paramlist types body) 
              (let* ([env_extendido (extender_env paramlist types ds ds)]
                                          [tipo_body (verificador body env_extendido)])
                                    (flecha (parseTiposAuxiliar (car types)) tipo_body))]
        [else (error "La función no pertenece al lenguaje MinTyRacket")]
        )]
    [APP (f_expr arglist)
         
         
         (if (VAR? f_expr) (Min)
             
             (let*
                 ([tipoT1 (verificador  f_expr  ds)]
                  [tipoT2 (verificador  (car arglist)  ds)]
                  ) 
               (cond  
                 [(Min? tipoT1) (Min)]
                 
                 [else (type-case tipo tipoT1
                         [flecha (antec consec) (if (esSubtipo tipoT2 
                                                               (flecha-antec tipoT1)) 
                                                    (flecha-consec tipoT1) 
                                                    (error "Parámetros de tipos inconsistentes"))]
                         [else (error "Tipo de funcion esperado")]
                         )])))])) 




#|Definición del ambiente de tipos|#
(define-type EnvTipos 
  [mtEnvTipos]
  [anEnvTipos (name symbol?) (value tipo_elem?) (env EnvTipos?)]
  [aRecSubTipos (name symbol?)
                (value isInMinTyRacket_Valores?)
                (env EnvTipos?)])




#|Procedimiento que busca el tipo recibido en un ambiente de tipos|#
(define (lookupTipo nombre ds)
  (type-case EnvTipos ds
    (mtEnvTipos () (error "no existe ligadura para el identificador: " nombre))
    (anEnvTipos (nombre_lig valor_lig env)
                (if (equal? nombre_lig nombre)
                    valor_lig
                    (lookupTipo nombre env)))
    [aRecSubTipos (nombrelig valorlig soc)
                  (if (symbol=? nombrelig nombre)
                      
                      (unbox valorlig)
                      (lookupTipo nombre soc))]))




#|Procedimiento que implementa las reglas de subtipificación|#
(define (esSubtipo sigma tau)
  (cond 
    [(Min? sigma) true]
    [(Max? tau) true]
    [(and       (booln? sigma)    (booln? tau)         ) true]
    [(and       (int? sigma)    (int? tau)        ) true]
    [(and       (float? sigma)    (float? tau)        ) true]
    [(and       (float? sigma)    (double? tau)        ) true]
    [(and       (int? sigma)    (double? tau)        ) true]
    [(and       (double? sigma)    (double? tau)        ) true]
    [(and       (int? sigma)    (float? tau)        ) true]
    [(and       (flecha? sigma)    (flecha? tau)     ) 
     (and (esSubtipo(flecha-antec tau) (flecha-antec sigma)) (esSubtipo(flecha-consec sigma) (flecha-consec tau)))]
    [(and       (lista? sigma)    (lista? tau)     )  (esSubtipo(lista-elementos sigma) (lista-elementos tau)) ]
    
    [else false]    
    ))



#|Definición de la lista compuesta por todos los tipos en MinTyRacket|#
(define listaTypes (list (int) (Min) (Max) (booln) (double) (float)))




#|Definición de la lista que almacena los supertipos comunes de dos tipos|#
(define supertiposComunes '())




#|Procedimiento que obtiene los supertipos comunes de dos tipos|#
(define (obtenerSupertiposComunes tipo_1 tipo_2 listaTipos supertiposComunes)
  (if (null? listaTipos)  supertiposComunes
      (cond 
        [(and (esSubtipo tipo_1 (car listaTipos)) (esSubtipo tipo_2 (car listaTipos))) 
            (obtenerSupertiposComunes tipo_1 tipo_2 (cdr listaTipos) (cons (car listaTipos) supertiposComunes))]
        
        [else (obtenerSupertiposComunes tipo_1 tipo_2 (cdr listaTipos) supertiposComunes)])))



(define listaTypes2 (list (int) (Min) (Max) (booln) (double) (float)))




#|Definición de la lista donde se almacenan todas las combinaciones de tipos definidos en MinTyRacket|#
(define combinacionTipos '())




#|Procedimiento que regresa todas las combinaciones de tipos definidos en MinTyRacket|#
(define (obtenerCombinaciones listaTipos1 listaTipos2 combinacionTipos)  
  (if (null? listaTipos1)
      combinacionTipos
      (obtenerCombinaciones 
       (cdr listaTipos1) listaTipos2 
       (cons (flecha (car listaTipos1)(car listaTipos2)) combinacionTipos))))


(define combinacionTipos2 '())
(define (obtenerTotalComb lista combinacionTipos)
  (if (null? lista)
      combinacionTipos
      (obtenerTotalComb (cdr lista) (append (obtenerCombinaciones listaTypes  lista combinacionTipos2) combinacionTipos))))



#|Procedimiento que obtiene las combinaciones de tipos posibles|#
(define combinacionTiposLista '())
(define (obtenerCombinacionesLista listaTipos1  combinacionTiposLista)  
  (if (null? listaTipos1)
      combinacionTiposLista
      (obtenerCombinacionesLista (cdr listaTipos1) (cons(lista (car listaTipos1)) combinacionTiposLista))))



#|Procedimiento que almacena todas las combinaciones de tipos posibles|#
(define TOTALTIPOS (append ( append (append (list (int) (Min) (Max) (booln) (double) (float))
                                            (obtenerCombinacionesLista listaTypes2 combinacionTiposLista) )
                                    (obtenerTotalComb listaTypes2 combinacionTipos))))




#|Procedimiento que calcula el mínimo supertipo común de dos tipos|#
(define (JOIN v_1 v_2)
  (let([listaSupertiposTotal (obtenerSupertiposComunes v_1 v_2 TOTALTIPOS supertiposComunes)])
    (elemMin listaSupertiposTotal(car  listaSupertiposTotal)) 
    ))




#|Procedimiento que calcula el tipo que es subtipo de los demás tipos en una lista|#
(define elemMin
  (lambda (ls min)
    
    (if (null? ls) min
        (elemMin (cdr ls) (minimo (car ls) min))
        )))




#|Procedimiento que regresa el subtipo entre dos tipos|#
(define minimo
  (lambda(ls1 ls2)
    (if (esSubtipo  ls1  ls2)
        ls1
        ls2 )))  




#|Procedimiento que aplica el algoritmo de tipifación dada una expresión escrita en sintaxis concreta|#
(define (VERIFICADOR exp envTipo)
  
  (cond 
    [(and (FIX? (abs exp))   (esSubtipo (verificador (abs exp) envTipo) (car (parseTipos exp))))
     (if
      (esSubtipo (flecha-consec  (verificador (abs exp) envTipo)) (second (parseTipos exp)))
      true
      false)]
    [else
     (if
      (esSubtipo (verificador (abs exp) envTipo) (parseTipos exp))
      true
      false
      )]))






